use std::any::TypeId;
use std::borrow::Cow;

#[derive(Debug,Clone,Copy)]
pub struct Descriptor {
	//in bytes
	pub offset: usize,
	//for type check
	pub tid: std::any::TypeId,
}
#[derive(Clone,Copy)]
pub enum Metadata{
	Has(fn(Cow<str>)->Option<&'static str>),
	No,
}

pub trait Reflector {
	// function returned there should check if struct has such field
	// internally, returned func should check existance of field on embedded match covering all field names.
	fn Reflect()-> (fn(Cow<str>)->Option<Descriptor>);
	fn Metadata()-> Metadata { Metadata::No }
}

#[derive(Debug,Clone,Copy)]
pub enum ReflErr{
	BadType,
	NoSuchField,
	NoSuchMetadata,
	NoMetadata,
}

pub fn get_meta<'a,T: Reflector + 'static,S: Into<Cow<'a,str>> >(name: S)->Result<&'static str,ReflErr>{
	use Metadata::*;
	match T::Metadata(){
		Has(f)=> {
			match f(name.into()){
				Some(x)=> Ok(x),
				None => Err(ReflErr::NoSuchMetadata),
			}
		},
		No => Err(ReflErr::NoMetadata),
	}

}

pub fn get<'a,T: Reflector + 'static,D: Copy + 'static, S: Into<Cow<'a,str>> >(trg: &'a mut T,name: S,data: &'a mut D)->Result<(),ReflErr>{ 
	match T::Reflect()(name.into()){
		Some(desc) => {
			if desc.tid == TypeId::of::<D>() {
				unsafe {
					let nptr = ((trg as *mut T as usize) + desc.offset) as *mut D;
					*data = *nptr;
				};
				return Ok(());
			} else {
				return Err(ReflErr::BadType);
			};
		},
		None => return Err(ReflErr::NoSuchField),
	}
}

pub fn set<'a,T: Reflector + 'static,D: Copy + 'static,S: Into<Cow<'a,str>> >(trg: &'a mut T,name: S,data: &'a mut D)->Result<(),ReflErr>{
	match T::Reflect()(name.into()){
		Some(desc) => {
			if desc.tid == TypeId::of::<D>() {
				unsafe {
					let nptr = ((trg as *mut T as usize) + desc.offset) as *mut D;
					*nptr=*data;
				};
				return Ok(());
			} else {
				return Err(ReflErr::BadType);
			};
		},
		None => return Err(ReflErr::NoSuchField),
	};
}

pub fn update<'a,T: Reflector + 'static,D: Copy + 'static,S: Into<Cow<'a,str>>,F: (Fn(&D)->D) >(trg: &'a mut T,name: S,f: F)->Result<(),ReflErr>{
	match T::Reflect()(name.into()){
		Some(desc) => {
			if desc.tid == TypeId::of::<D>() {
				unsafe {
					let nptr = ((trg as *mut T as usize) + desc.offset) as *mut D;
					*nptr=f({
						match nptr.as_ref(){
							Some(ret) =>ret,
							None => unreachable!(),
						}
					});
				};
				return Ok(());
			} else {
				return Err(ReflErr::BadType);
			};
		},
		None => return Err(ReflErr::NoSuchField),
	};
}
